import json
import logging
import sys

import click

from biolib._index.index import Index
from biolib.biolib_errors import BioLibError
from biolib.biolib_logging import logger, logger_no_user_data


@click.group(help='Manage Indexes')
def index() -> None:
    logger.configure(default_log_level=logging.INFO)
    logger_no_user_data.configure(default_log_level=logging.INFO)


@index.command(help='Create an Index')
@click.argument('uri', required=True)
@click.option('--config-path', required=True, type=click.Path(exists=True), help='Path to JSON config file')
def create(uri: str, config_path: str) -> None:
    try:
        Index.create_from_config_file(uri=uri, config_path=config_path)
    except json.JSONDecodeError as error:
        print(f'Error: Invalid JSON in config file: {error}', file=sys.stderr)
        sys.exit(1)
    except BioLibError as error:
        print(f'Error creating index: {error.message}', file=sys.stderr)
        sys.exit(1)
    except Exception as error:
        print(f'Error reading config file: {error}', file=sys.stderr)
        sys.exit(1)



import os
from collections import namedtuple
from datetime import datetime
from pathlib import Path
from struct import Struct
from typing import Callable, Dict, Iterable, Iterator, List, Optional, Union, cast

from biolib import api
from biolib._internal.data_record import get_data_record_state_from_uri
from biolib._internal.data_record.data_record import validate_sqlite_v1
from biolib._internal.data_record.push_data import (
    _upload_from_iterator,
    push_data_path,
    validate_data_path_and_get_files_and_size_of_directory,
)
from biolib._internal.data_record.remote_storage_endpoint import DataRecordRemoteStorageEndpoint
from biolib._internal.http_client import HttpClient
from biolib._shared import types
from biolib._shared.types import ZipFileNodeDict
from biolib._shared.utils import parse_resource_uri
from biolib.api import client as api_client
from biolib.biolib_api_client import BiolibApiClient
from biolib.biolib_api_client.lfs_types import DataRecordInfo, DataRecordVersionInfo
from biolib.biolib_binary_format import LazyLoadedFile
from biolib.biolib_binary_format.utils import RemoteIndexableBuffer
from biolib.biolib_logging import logger

PathFilter = Union[str, List[str], Callable[[str], bool]]


class DataRecord:
    def __init__(self, _internal_state: DataRecordVersionInfo):
        self._state = _internal_state

    def __repr__(self):
        return f'DataRecord: {self._state["resource_uri"]}'

    @property
    def uri(self) -> str:
        return self._state['resource_uri']

    @property
    def uuid(self) -> str:
        return self._state['resource_uuid']

    @property
    def name(self) -> str:
        uri_parsed = parse_resource_uri(self._state['resource_uri'], use_account_as_name_default=False)
        if not uri_parsed['resource_name']:
            raise ValueError('Expected parameter "resource_uri" to contain resource name')

        return uri_parsed['resource_name']

    def list_files(
        self,
        path_filter: Optional[PathFilter] = None,
        max_count: Optional[int] = 100_000,
    ) -> List[LazyLoadedFile]:
        files = list(
            self._fetch_files(
                path_filter=path_filter,
                max_count=max_count + 1 if max_count is not None else None,
            )
        )

        if max_count is not None and len(files) > max_count:
            raise Exception(
                f'list_files returned more than {max_count} files. '
                f'Please set the keyword argument "max_count" to a higher number.'
            )

        return files

    def download_zip(self, output_path: str):
        remote_storage_endpoint = DataRecordRemoteStorageEndpoint(
            resource_version_uuid=self._state['resource_version_uuid'],
        )
        HttpClient.request(url=remote_storage_endpoint.get_remote_url(), response_path=output_path)

    def download_files(self, output_dir: str, path_filter: Optional[PathFilter] = None) -> None:
        filtered_files = self.list_files(path_filter=path_filter)

        if len(filtered_files) == 0:
            logger.debug('No files to save')
            return

        for file in filtered_files:
            file_path = os.path.join(output_dir, file.path)
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(file_path, mode='wb') as file_handle:
                for chunk in file.get_data_iterator():
                    file_handle.write(chunk)

    def save_files(self, output_dir: str, path_filter: Optional[PathFilter] = None) -> None:
        self.download_files(output_dir=output_dir, path_filter=path_filter)

    def update(self, data_path: str, chunk_size_in_mb: Optional[int] = None) -> None:
        BiolibApiClient.assert_is_signed_in(authenticated_action_description='push data to a Data Record')
        files_to_zip, data_size_in_bytes = validate_data_path_and_get_files_and_size_of_directory(data_path)

        # validate data record
        detailed_dict: types.DataRecordDetailedDict = self._get_detailed_dict()
        if detailed_dict['type']:
            # only validate if data record has a type
            data_record_type: types.DataRecordTypeDict = detailed_dict['type']
            logger.info(f"Validating data record of type {data_record_type['name']}")
            for rule in data_record_type['validation_rules']:
                logger.info(f"Validating rule {rule['type']} for {rule['path']}...")
                if rule['type'] == 'sqlite-v1':
                    try:
                        validate_sqlite_v1(schema=rule['rule'], sqlite_file=Path(rule['path']))
                    except Exception as error:
                        raise Exception('Data Record Validation failed') from error
                else:
                    raise Exception(f"Error processing data record validation: unknown rule type {rule['type']}")

        new_resource_version_uuid = push_data_path(
            data_path=data_path,
            data_size_in_bytes=data_size_in_bytes,
            files_to_zip=files_to_zip,
            resource_uuid=self._state['resource_uuid'],
            chunk_size_in_mb=chunk_size_in_mb,
            publish=True,
        )

        updated_record = DataRecord._get_by_version_uuid(new_resource_version_uuid)
        self._state = updated_record._state  # pylint: disable=protected-access
        logger.info(f"Successfully pushed a new Data Record version '{self.uri}'")

    @staticmethod
    def get_by_uri(uri: str) -> 'DataRecord':
        return DataRecord(_internal_state=get_data_record_state_from_uri(uri))

    @staticmethod
    def _get_by_version_uuid(version_uuid: str) -> 'DataRecord':
        response = api.client.get(path=f'/lfs/versions/{version_uuid}/')
        version_info = response.json()
        return DataRecord.get_by_uri(version_info['uri'])

    @staticmethod
    def create(destination: str, data_path: Optional[str] = None, record_type: Optional[str] = None) -> 'DataRecord':
        BiolibApiClient.assert_is_signed_in(authenticated_action_description='create a Data Record')
        if data_path is not None:
            assert os.path.isdir(data_path), f'The path "{data_path}" is not a directory.'
        uri_parsed = parse_resource_uri(destination, use_account_as_name_default=False)
        if uri_parsed['resource_name_normalized']:
            data_record_uri = destination
        else:
            record_name = 'data-record-' + datetime.now().isoformat().split('.')[0].replace(':', '-')
            data_record_uri = f'{destination}/{record_name}'

        response = api.client.post(
            path='/resources/data-records/',
            data={
                'uri': data_record_uri,
                'type': record_type,
            },
        )
        data_record_info: DataRecordInfo = response.json()
        logger.info(f"Successfully created new Data Record '{data_record_info['uri']}'")

        data_record = DataRecord.get_by_uri(uri=data_record_info['uri'])
        if data_path is not None:
            data_record.update(data_path=data_path)

        return data_record

    @staticmethod
    def fetch(uri: Optional[str] = None, count: Optional[int] = None) -> List['DataRecord']:
        max_page_size = 1_000
        params: Dict[str, Union[str, int]] = {
            'page_size': str(count or max_page_size),
            'resource_type': 'data-record',
        }
        if uri:
            uri_parsed = parse_resource_uri(uri, use_account_as_name_default=False)
            params['account_handle'] = uri_parsed['account_handle_normalized']
            if uri_parsed['resource_name_normalized']:
                params['app_name'] = uri_parsed['resource_name_normalized']

        results = api_client.get(path='/apps/', params=params).json()['results']
        if count is None and len(results) == max_page_size:
            logger.warning(
                f'Fetch results exceeded maximum count of {max_page_size}. Some data records might not be fetched.'
            )

        return [
            DataRecord(
                _internal_state={
                    'resource_uri': result['resource_uri'],
                    'resource_uuid': result['public_id'],
                    'resource_version_uuid': result['active_version'],
                }
            )
            for result in results
        ]

    def _fetch_files(
        self,
        max_count: Optional[int],
        path_filter: Optional[PathFilter] = None,
    ) -> Iterable[LazyLoadedFile]:
        if path_filter and not (isinstance(path_filter, (str, list)) or callable(path_filter)):
            raise Exception('Expected path_filter to be a string, a list of strings or a function')

        path_filters = (
            [path_filter] if isinstance(path_filter, str) else path_filter if isinstance(path_filter, list) else []
        )

        resource_version_uuid = self._state['resource_version_uuid']
        remote_storage_endpoint = DataRecordRemoteStorageEndpoint(resource_version_uuid)

        page: Optional[int] = 1
        yielded_files: int = 0
        while page:
            response = api.client.post(
                path=f'/proxy/files/data-record-versions/{resource_version_uuid}/query/',
                data=dict(page=page, page_size=1_000, path_filters=path_filters),
            ).json()

            for file_node_dict in cast(List[ZipFileNodeDict], response['results']):
                if file_node_dict['is_dir']:
                    continue

                if callable(path_filter) and not path_filter(file_node_dict['dir_path'] + file_node_dict['name']):
                    continue

                yield self._get_file(remote_storage_endpoint, file_node_dict)
                yielded_files += 1

                if max_count is not None and yielded_files >= max_count:
                    page = None
                    break

            page = page + 1 if page is not None and response['page_count'] > page else None

    @staticmethod
    def _get_file(
        remote_storage_endpoint: DataRecordRemoteStorageEndpoint,
        file_node_dict: ZipFileNodeDict,
    ) -> LazyLoadedFile:
        local_file_header_signature_bytes = b'\x50\x4b\x03\x04'
        local_file_header_struct = Struct('<H2sHHHIIIHH')
        LocalFileHeader = namedtuple(
            'LocalFileHeader',
            (
                'version',
                'flags',
                'compression_raw',
                'mod_time',
                'mod_date',
                'crc_32_expected',
                'compressed_size_raw',
                'uncompressed_size_raw',
                'file_name_len',
                'extra_field_len',
            ),
        )

        local_file_header_start = file_node_dict['zip_meta']['header_start'] + len(local_file_header_signature_bytes)
        local_file_header_end = local_file_header_start + local_file_header_struct.size

        def file_start_func() -> int:
            local_file_header_response = HttpClient.request(
                url=remote_storage_endpoint.get_remote_url(),
                headers={'range': f'bytes={local_file_header_start}-{local_file_header_end - 1}'},
                timeout_in_seconds=300,
            )
            local_file_header = LocalFileHeader._make(
                local_file_header_struct.unpack(local_file_header_response.content)
            )
            file_start: int = (
                local_file_header_end + local_file_header.file_name_len + local_file_header.extra_field_len
            )
            return file_start

        return LazyLoadedFile(
            buffer=RemoteIndexableBuffer(endpoint=remote_storage_endpoint),
            length=file_node_dict['zip_meta']['size_on_disk'],
            path=file_node_dict['dir_path'] + file_node_dict['name'],
            start=None,
            start_func=file_start_func,
        )

    def _get_detailed_dict(self) -> types.DataRecordDetailedDict:
        return cast(types.DataRecordDetailedDict, api_client.get(f'/resources/data-records/{self.uuid}/').json())

    def _get_zip_size_bytes(self) -> int:
        remote_storage_endpoint = DataRecordRemoteStorageEndpoint(
            resource_version_uuid=self._state['resource_version_uuid'],
        )
        presigned_url = remote_storage_endpoint.get_remote_url()
        response = HttpClient.request(url=presigned_url, headers={'range': 'bytes=0-0'})
        content_range = response.headers.get('Content-Range', '')
        if not content_range or '/' not in content_range:
            raise ValueError('Unable to determine zip size: Content-Range header missing or invalid')
        total_size = int(content_range.split('/')[1])
        return total_size

    def _iter_zip_bytes(self, chunk_size_bytes: int) -> Iterator[bytes]:
        remote_storage_endpoint = DataRecordRemoteStorageEndpoint(
            resource_version_uuid=self._state['resource_version_uuid'],
        )
        presigned_url = remote_storage_endpoint.get_remote_url()
        response = HttpClient.request(url=presigned_url, headers={'range': 'bytes=0-0'})
        content_range = response.headers.get('Content-Range', '')
        if not content_range or '/' not in content_range:
            raise ValueError('Unable to determine zip size: Content-Range header missing or invalid')
        total_size = int(content_range.split('/')[1])

        for start in range(0, total_size, chunk_size_bytes):
            end = min(start + chunk_size_bytes - 1, total_size - 1)
            presigned_url = remote_storage_endpoint.get_remote_url()
            response = HttpClient.request(
                url=presigned_url,
                headers={'range': f'bytes={start}-{end}'},
                timeout_in_seconds=300,
            )
            yield response.content

    @staticmethod
    def clone(
        source: 'DataRecord',
        destination: 'DataRecord',
        on_progress: Optional[Callable[[int, int], None]] = None,
    ) -> 'DataRecord':
        BiolibApiClient.assert_is_signed_in(authenticated_action_description='clone a Data Record')

        # pylint: disable=protected-access
        total_size_in_bytes = source._get_zip_size_bytes()

        if total_size_in_bytes == 0:
            raise ValueError('Source data record has no data to clone')

        min_chunk_size_bytes = 10_000_000
        chunk_size_in_bytes = max(min_chunk_size_bytes, int(total_size_in_bytes / 9_000))

        zip_iterator = source._iter_zip_bytes(chunk_size_bytes=chunk_size_in_bytes)

        new_resource_version_uuid = _upload_from_iterator(
            resource_uuid=destination._state['resource_uuid'],
            payload_iterator=zip_iterator,
            payload_size_in_bytes=total_size_in_bytes,
            publish=True,
            on_progress=on_progress,
        )
        # pylint: enable=protected-access

        logger.info(f"Successfully cloned data to '{destination.uri}'")
        return DataRecord._get_by_version_uuid(new_resource_version_uuid)

    def delete(self) -> None:
        """Delete the data record.

        Example::
            >>> record = DataRecord.get_by_uri("account/data-record")
            >>> record.delete()
        """
        try:
            api_client.delete(path=f'/apps/{self.uuid}/')
            logger.info(f'Data record {self.uri} deleted')
        except Exception as error:
            raise Exception(f'Failed to delete data record {self.uri} due to: {error}') from error



from .typing import Literal, NotRequired, Optional, TypedDict


class ResourceVersionAssetsDict(TypedDict):
    download_url: str
    size_bytes: int


class ResourceVersionDict(TypedDict):
    uuid: str
    semantic_version: str
    state: Literal['published', 'unpublished']
    created_at: str
    git_branch_name: NotRequired[str]
    git_commit_hash: NotRequired[str]


class ResourceVersionDetailedDict(ResourceVersionDict):
    assets: Optional[ResourceVersionAssetsDict]



import base64
import binascii
import json
from typing import Any, Dict


class JwtDecodeError(Exception):
    pass


def decode_jwt_without_checking_signature(jwt: str) -> Dict[str, Any]:
    jwt_bytes = jwt.encode('utf-8')

    try:
        signing_input, _ = jwt_bytes.rsplit(b'.', 1)
        header_segment, payload_segment = signing_input.split(b'.', 1)
    except ValueError as error:
        raise JwtDecodeError('Not enough segments') from error

    try:
        header_data = base64.urlsafe_b64decode(header_segment)
    except (TypeError, binascii.Error) as error:
        raise JwtDecodeError('Invalid header padding') from error

    try:
        header = json.loads(header_data)
    except ValueError as error:
        raise JwtDecodeError(f'Invalid header string: {error}') from error

    if not isinstance(header, dict):
        raise JwtDecodeError('Invalid header string: must be a json object')

    try:
        payload_data = base64.urlsafe_b64decode(payload_segment)
    except (TypeError, binascii.Error) as error:
        raise JwtDecodeError('Invalid payload padding') from error

    try:
        payload = json.loads(payload_data)
    except ValueError as error:
        raise JwtDecodeError(f'Invalid payload string: {error}') from error

    if not isinstance(payload, dict):
        raise JwtDecodeError('Invalid payload string: must be a json object')

    return dict(header=header, payload=payload)



import mimetypes
import os
import random
import re
import subprocess
import urllib.parse

import biolib.api
from biolib import biolib_errors
from biolib._internal.http_client import HttpError
from biolib.api.client import ApiClient
from biolib.biolib_api_client import AppGetResponse
from biolib.biolib_logging import logger
from biolib.typing_utils import Optional
from biolib.utils import load_base_url_from_env


def encode_multipart(data, files):
    boundary = f'----------{random.randint(0, 1000000000)}'
    line_array = []

    for key, value in data.items():
        if value is not None:
            line_array.append(f'--{boundary}')
            line_array.append(f'Content-Disposition: form-data; name="{key}"')
            line_array.append('')
            line_array.append(value)

    for key, (filename, value) in files.items():
        line_array.append(f'--{boundary}')
        line_array.append(f'Content-Disposition: form-data; name="{key}"; filename="{filename}"')
        line_array.append(f'Content-Type: {mimetypes.guess_type(filename)[0] or "application/octet-stream"}')
        line_array.append('')
        line_array.append('')
        line_array.append(value)

    line_array.append(f'--{boundary}--')
    line_array.append('')

    data_encoded = b'\r\n'.join([line.encode() if isinstance(line, str) else line for line in line_array])
    return f'multipart/form-data; boundary={boundary}', data_encoded


def _get_git_branch_name() -> str:
    try:
        github_actions_branch_name = os.getenv('GITHUB_REF_NAME')
        if github_actions_branch_name:
            return github_actions_branch_name

        gitlab_ci_branch_name = os.getenv('CI_COMMIT_REF_NAME')
        if gitlab_ci_branch_name:
            return gitlab_ci_branch_name

        result = subprocess.run(['git', 'branch', '--show-current'], check=True, stdout=subprocess.PIPE, text=True)
        return result.stdout.strip()
    except BaseException:
        return ''


def _get_git_commit_hash() -> str:
    try:
        github_actions_commit_hash = os.getenv('GITHUB_SHA')
        if github_actions_commit_hash:
            return github_actions_commit_hash

        gitlab_ci_commit_hash = os.getenv('CI_COMMIT_SHA')
        if gitlab_ci_commit_hash:
            return gitlab_ci_commit_hash

        result = subprocess.run(['git', 'rev-parse', 'HEAD'], check=True, stdout=subprocess.PIPE, text=True)
        return result.stdout.strip()
    except BaseException:
        return ''


def _get_git_repository_url() -> str:
    try:
        result = subprocess.run(['git', 'remote', 'get-url', 'origin'], check=True, stdout=subprocess.PIPE, text=True)
        return result.stdout.strip()
    except BaseException:
        return ''

def _get_app_uri_from_str(input_str: str) -> str:
    parsed_base_url = urllib.parse.urlparse(load_base_url_from_env())
    parsed_uri = urllib.parse.urlparse(input_str)
    if parsed_uri.netloc != '' and parsed_base_url.netloc != parsed_uri.netloc:
        raise biolib_errors.ValidationError(
            f'Invalid URI. The hostname "{parsed_base_url.netloc}" is not recognized.'
        )
    elif parsed_uri.netloc != '' and parsed_uri.path[1] != '@':
        uri = f'@{parsed_uri.netloc}{parsed_uri.path}'
    elif parsed_uri.netloc == '' and parsed_uri.path.startswith (parsed_base_url.netloc):
        uri = f'@{parsed_uri.path}'
    else:
        uri = parsed_uri.path
    uri = uri.strip('/')
    # Replace frontend version path with app_uri compatible version (if supplied)
    uri = re.sub(r'/version/(?P<version>(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*))(/)?$', r':\g<version>', uri)
    return uri


class BiolibAppApi:
    @staticmethod
    def get_by_uri(uri: str, api_client: Optional[ApiClient] = None) -> AppGetResponse:
        uri = _get_app_uri_from_str(uri)
        api = api_client or biolib.api.client
        try:
            response = api.get(path='/app/', params={'uri': uri})
            app_response: AppGetResponse = response.json()
            return app_response

        except HttpError as error:
            if error.code == 404:
                raise biolib_errors.NotFound(f'Application {uri} not found.') from None

            raise error

    @staticmethod
    def create_app(uri: str):
        uri = _get_app_uri_from_str(uri)
        try:
            response = biolib.api.client.post(path='/resources/apps/', data={'uri': uri})
            return response.json()
        except HttpError as error:
            raise error

    @staticmethod
    def push_app_version(
        app_id,
        zip_binary,
        author,
        app_name,
        set_as_active,
        app_version_id_to_copy_images_from: Optional[str],
        semantic_version: Optional[str],
    ):
        try:
            data = {
                'app': app_id,
                'set_as_active': 'true' if set_as_active else 'false',
                'state': 'published',
                'app_version_id_to_copy_images_from': app_version_id_to_copy_images_from,
                'git_branch_name': _get_git_branch_name(),
                'git_commit_hash': _get_git_commit_hash(),
                'git_repository_url': _get_git_repository_url(),
            }
            if semantic_version:
                data['semantic_version'] = semantic_version

            content_type, data_encoded = encode_multipart(
                data=data,
                files={
                    'source_files_zip': ('source_files.zip', zip_binary),
                },
            )
            response = biolib.api.client.post(
                path='/app_versions/',
                data=data_encoded,
                headers={'Content-Type': content_type},
            )
        except Exception as error:
            logger.error(f'Push failed for {author}/{app_name}:')
            raise error

        # TODO: When response includes the version number, print the URL for the new app version
        logger.info(f'Initialized new app version for {author}/{app_name}.')
        return response.json()



from biolib._shared.utils.resource_uri import normalize_resource_name, parse_resource_uri, parse_semantic_version

__all__ = [
    'normalize_resource_name',
    'parse_resource_uri',
    'parse_semantic_version',
]



import os
import shutil
import sys

import click

from biolib import (
    biolib_errors,
    utils,  # Import like this to let BASE_URL_IS_PUBLIC_BIOLIB be set correctly
)
from biolib._internal.add_copilot_prompts import add_copilot_prompts
from biolib._internal.add_gui_files import add_gui_files
from biolib._internal.http_client import HttpClient, HttpError
from biolib._internal.string_utils import normalize_for_docker_tag
from biolib._internal.templates import templates
from biolib.api import client as api_client
from biolib.biolib_api_client.api_client import BiolibApiClient
from biolib.biolib_api_client.biolib_app_api import BiolibAppApi
from biolib.biolib_logging import logger_no_user_data
from biolib.typing_utils import Optional
from biolib.user.sign_in import sign_in
from biolib.utils import BIOLIB_PACKAGE_VERSION


def _get_latest_pypi_version() -> Optional[str]:
    try:
        response = HttpClient.request(
            url='https://pypi.org/pypi/pybiolib/json',
            timeout_in_seconds=5,
            retries=1,
        )
        data = response.json()
        version = data.get('info', {}).get('version')
        if isinstance(version, str):
            return version
        return None
    except Exception as error:
        logger_no_user_data.debug(f'Failed to fetch latest version from PyPI: {error}')
        return None


def _is_current_version_outdated(current: str, latest: str) -> bool:
    try:
        current_parts = [int(x) for x in current.split('.')]
        latest_parts = [int(x) for x in latest.split('.')]
        return current_parts < latest_parts
    except (ValueError, AttributeError):
        return False


@click.command(help='Initialize a BioLib project', hidden=True)
def init() -> None:
    latest_version = _get_latest_pypi_version()
    if latest_version and _is_current_version_outdated(BIOLIB_PACKAGE_VERSION, latest_version):
        print(f'A newer version of pybiolib is available: {latest_version} (current: {BIOLIB_PACKAGE_VERSION})')
        print('To upgrade, run: pip install --upgrade pybiolib')
        print()
        continue_input = input('Do you want to continue with the current version? [y/N]: ')
        if continue_input.lower() not in ['y', 'yes']:
            print('Please upgrade pybiolib and run `biolib init` again.')
            return

    cwd = os.getcwd()

    app_uri = input('What URI do you want to create the application under? (leave blank to skip): ')

    if app_uri and not app_uri.startswith('@'):
        try:
            response = api_client.get('system/enterprise/config/', authenticate=False)
            config = response.json()
            prefix = config.get('resource_hostname_prefix')
            if prefix:
                app_uri = f'@{prefix}/{app_uri}'
                print(f'Detected enterprise deployment, using URI: {app_uri}')
        except HttpError as e:
            # 404 indicates endpoint not found, 501 indicates non-enterprise deployment
            if e.code in [404, 501]:
                pass
            else:
                print(f'Warning: Could not detect enterprise configuration: {e}')
        except Exception as e:
            print(f'Warning: Could not detect enterprise configuration: {e}')

    app_name = app_uri.split('/')[-1] if app_uri else None
    docker_tag = normalize_for_docker_tag(app_name) if app_name else None

    if app_uri:
        try:
            if BiolibApiClient.is_reauthentication_needed():
                sign_in_input = input('You need to sign in to validate/create apps. Would you like to sign in? [y/N]: ')
                if sign_in_input.lower() in ['y', 'yes']:
                    sign_in()
                else:
                    print('Skipping app validation and creation. You can set the URI in .biolib/config.yml later.')
                    return

            BiolibAppApi.get_by_uri(app_uri)
            print(f'App {app_uri} already exists.')
        except biolib_errors.NotFound:
            create_app_input = input(f'App {app_uri} does not exist. Would you like to create it? [y/N]: ')
            if create_app_input.lower() in ['y', 'yes']:
                try:
                    BiolibAppApi.create_app(app_uri)
                    print(f'Successfully created app {app_uri}')
                except Exception as e:
                    print(f'Failed to create app {app_uri}: {str(e)}')
                    print('You can create the app manually later or set the URI in .biolib/config.yml')
            else:
                print(
                    'App creation skipped. You can create the app manually later or set the URI in .biolib/config.yml'
                )
        except Exception as e:
            print(f'Failed to validate app {app_uri}: {str(e)}')
            print('Continuing with initialization...')

    if not app_uri:
        print(
            'Remember to set the app URI in the .biolib/config.yml file later, '
            'and docker image name in the .biolib/config.yml and .github/workflows/biolib.yml files.'
        )
    advanced_setup_input = input('Do you want to set up advanced features like Copilot and GUI? [y/N]: ')
    advanced_setup = advanced_setup_input.lower() == 'y'
    include_copilot = False
    include_gui = False
    if advanced_setup:
        copilot_enabled_input = input('Do you want to include Copilot instructions and prompts? [y/N]: ')
        include_copilot = copilot_enabled_input.lower() == 'y'
        include_gui_input = input('Do you want to include GUI setup? [y/N]: ')
        include_gui = include_gui_input.lower() == 'y'

    init_template_dir = templates.init_template()
    conflicting_files = []
    files_to_overwrite = set()

    try:
        # First pass: check for conflicts
        for root, dirs, filenames in os.walk(init_template_dir):
            dirs[:] = [d for d in dirs if '__pycache__' not in d]
            relative_dir = os.path.relpath(root, init_template_dir)
            destination_dir = cwd if relative_dir == '.' else os.path.join(cwd, relative_dir)

            for filename in filenames:
                source_file = os.path.join(root, filename)
                destination_file = os.path.join(destination_dir, filename)
                if os.path.exists(destination_file):
                    with open(source_file, 'rb') as fsrc, open(destination_file, 'rb') as fdest:
                        if fsrc.read() != fdest.read():
                            conflicting_files.append(os.path.relpath(destination_file, cwd))

        if conflicting_files:
            print('The following files already exist and would be overwritten:')
            for conflicting_file in conflicting_files:
                print(f'  {conflicting_file}')
            print()

            for conflicting_file in conflicting_files:
                choice = input(f'Overwrite {conflicting_file}? [y/N]: ').lower().strip()
                if choice in ['y', 'yes']:
                    files_to_overwrite.add(conflicting_file)

        replace_app_uri = app_uri if app_uri else 'PUT_APP_URI_HERE'
        replace_app_name = app_name if app_name else 'biolib-app'

        # Second pass: copy files (only if no conflicts)
        for root, dirs, filenames in os.walk(init_template_dir):
            dirs[:] = [d for d in dirs if '__pycache__' not in d]
            relative_dir = os.path.relpath(root, init_template_dir)
            destination_dir = os.path.join(cwd, relative_dir)

            os.makedirs(destination_dir, exist_ok=True)

            for filename in filenames:
                if utils.BASE_URL_IS_PUBLIC_BIOLIB and filename == 'biolib.yml':
                    continue

                relative_file_path = os.path.join(relative_dir, filename) if relative_dir != '.' else filename

                source_file = os.path.join(root, filename)
                destination_file = os.path.join(destination_dir, filename)
                relative_file_path = os.path.relpath(destination_file, cwd)

                if not os.path.exists(destination_file) or relative_file_path in files_to_overwrite:
                    try:
                        with open(source_file) as f:
                            content = f.read()

                        new_content = content.replace('BIOLIB_REPLACE_PYBIOLIB_VERSION', BIOLIB_PACKAGE_VERSION)
                        new_content = new_content.replace('BIOLIB_REPLACE_APP_URI', replace_app_uri)
                        new_content = new_content.replace(
                            'BIOLIB_REPLACE_DOCKER_TAG',
                            docker_tag if docker_tag else 'PUT_DOCKER_TAG_HERE',
                        )
                        new_content = new_content.replace('BIOLIB_REPLACE_APP_NAME', replace_app_name)

                        gui_config = "main_output_file: '/result.html'\n" if include_gui else ''
                        new_content = new_content.replace('BIOLIB_REPLACE_GUI_CONFIG\n', gui_config)

                        gui_mv_command = 'mv result.html output/result.html\n' if include_gui else ''
                        new_content = new_content.replace('BIOLIB_REPLACE_GUI_MV_COMMAND\n', gui_mv_command)

                        with open(destination_file, 'w') as f:
                            f.write(new_content)
                    except UnicodeDecodeError:
                        shutil.copy2(source_file, destination_file)

        readme_path = os.path.join(cwd, 'README.md')
        if not os.path.exists(readme_path) and app_name:
            with open(readme_path, 'w') as readme_file:
                readme_file.write(f'# {app_name}\n')

        if include_copilot:
            add_copilot_prompts(force=False, silent=True)

        if include_gui:
            add_gui_files(force=False, silent=True)

    except KeyboardInterrupt:
        print('\nInit command cancelled.', file=sys.stderr)
        exit(1)



import os
import shutil
import sys

from biolib._internal.templates import templates


def add_gui_files(force=False, silent=False) -> None:
    cwd = os.getcwd()
    template_dir = templates.gui_template()

    root_files = ['package.json', 'Dockerfile', 'index.html', 'vite.config.mts', '.yarnrc.yml']

    conflicting_files = []

    # Copy root_files to init_template root and rest to subdirectory
    for root, _, filenames in os.walk(template_dir):
        relative_dir = os.path.relpath(root, template_dir)

        for filename in filenames:
            if filename in root_files:
                destination_dir = cwd
            else:
                if relative_dir == '.':
                    destination_dir = os.path.join(cwd, 'gui')
                else:
                    destination_dir = os.path.join(cwd, 'gui', relative_dir)

            source_file = os.path.join(root, filename)
            destination_file = os.path.join(destination_dir, filename)

            if filename == 'Dockerfile':
                force = True

            if os.path.exists(destination_file) and not force:
                with open(source_file, 'rb') as fsrc, open(destination_file, 'rb') as fdest:
                    if fsrc.read() != fdest.read():
                        conflicting_files.append(os.path.relpath(destination_file, cwd))
            else:
                os.makedirs(destination_dir, exist_ok=True)
                shutil.copy2(source_file, destination_file)

    gitignore_path = os.path.join(cwd, '.gitignore')
    with open(gitignore_path, 'a') as gitignore_file:
        gitignore_file.write('\n# gui\n')
        gitignore_file.write('.yarn\n')
        gitignore_file.write('dist\n')
        gitignore_file.write('yarn.lock\n')
        gitignore_file.write('tsconfig.tsbuildinfo\n')
        gitignore_file.write('node_modules\n')

    if conflicting_files:
        print('The following files were not overwritten. Use --force to override them:', file=sys.stderr)
        for conflicting_file in list(set(conflicting_files)):
            print(f'  {conflicting_file}', file=sys.stderr)
        exit(1)

    if not silent:
        print('gui files added to project root and gui/ subdirectory')



import http.client
import json
import platform
import shutil
import socket
import ssl
import subprocess
import time
import urllib.error
import urllib.parse
import urllib.request

from biolib.biolib_logging import logger_no_user_data
from biolib.typing_utils import Dict, Literal, Optional, Union, cast

_HttpMethod = Literal['GET', 'POST', 'PATCH', 'PUT', 'DELETE']


def _create_ssl_context():
    context = ssl.create_default_context()
    try:
        if platform.system() == 'Darwin':
            certificates = subprocess.check_output('security find-certificate -a -p', shell=True).decode('utf-8')
            context.load_verify_locations(cadata=certificates)
    except BaseException:
        pass
    return context


class HttpError(urllib.error.HTTPError):
    def __init__(self, http_error: urllib.error.HTTPError):
        super().__init__(
            url=http_error.url,
            code=http_error.code,
            msg=http_error.msg,  # type: ignore
            hdrs=http_error.hdrs,  # type: ignore
            fp=http_error.fp,
        )

    def __str__(self):
        response_text = self.read().decode('utf-8')
        return f'{self.code} Error: {response_text} for url: {self.url}'


class HttpResponse:
    def __init__(self, response, response_path) -> None:
        self.headers: Dict[str, str] = dict(response.headers)
        self.status_code: int = int(response.status)
        self.response_path = response_path
        if self.response_path:
            with open(self.response_path, 'wb') as out_file:
                shutil.copyfileobj(response, out_file)
        else:
            self.content: bytes = response.read()
        self.url: str = response.geturl()

    @property
    def text(self) -> str:
        if self.response_path:
            with open(self.response_path, 'rb') as fp:
                return cast(str, fp.read().decode('utf-8'))
        else:
            return cast(str, self.content.decode('utf-8'))

    def json(self):
        return json.loads(self.text)


class HttpClient:
    ssl_context = None

    @staticmethod
    def request(
        url: str,
        method: Optional[_HttpMethod] = None,
        data: Optional[Union[Dict, bytes]] = None,
        headers: Optional[Dict[str, str]] = None,
        retries: int = 5,
        timeout_in_seconds: Optional[int] = None,
        response_path: Optional[str] = None,
        retry_on_http_500: Optional[bool] = False,
        max_content_length_in_bytes: Optional[int] = None,
    ) -> HttpResponse:
        if not HttpClient.ssl_context:
            HttpClient.ssl_context = _create_ssl_context()
        headers_to_send = headers or {}
        if isinstance(data, dict):
            headers_to_send['Accept'] = 'application/json'
            headers_to_send['Content-Type'] = 'application/json'

        request = urllib.request.Request(
            url=url,
            data=json.dumps(data).encode() if isinstance(data, dict) else data,
            headers=headers_to_send,
            method=method or 'GET',
        )
        if timeout_in_seconds is None:
            timeout_in_seconds = 60 if isinstance(data, dict) else 180  # TODO: Calculate timeout based on data size

        last_error: Optional[Exception] = None
        for retry_count in range(retries + 1):
            if retry_count > 0:
                time.sleep(5 * retry_count)
                logger_no_user_data.debug(f'Retrying HTTP {method} request...')
            try:
                with urllib.request.urlopen(
                    request,
                    context=HttpClient.ssl_context,
                    timeout=timeout_in_seconds,
                ) as response:
                    if max_content_length_in_bytes:
                        content_length = response.getheader('Content-Length')
                        if not content_length:
                            raise ValueError('No Content-Length in response headers')

                        if int(content_length) > max_content_length_in_bytes:
                            raise ValueError(f'Content-Length exceeds {max_content_length_in_bytes} bytes')

                    return HttpResponse(response, response_path)

            except urllib.error.HTTPError as error:
                if error.code == 429:
                    logger_no_user_data.warning(f'HTTP {method} request failed with status 429 for "{url}"')
                    last_error = error
                elif error.code == 500 and retry_on_http_500:
                    logger_no_user_data.warning(f'HTTP {method} request failed with status 500 for "{url}"')
                    last_error = error
                elif error.code == 502:
                    logger_no_user_data.warning(f'HTTP {method} request failed with status 502 for "{url}"')
                    last_error = error
                elif error.code == 503:
                    logger_no_user_data.warning(f'HTTP {method} request failed with status 503 for "{url}"')
                    last_error = error
                elif error.code == 504:
                    logger_no_user_data.warning(f'HTTP {method} request failed with status 504 for "{url}"')
                    last_error = error
                else:
                    raise HttpError(error) from None

            except urllib.error.URLError as error:
                if isinstance(error.reason, socket.timeout):
                    if retry_count > 0:
                        logger_no_user_data.warning(f'HTTP {method} request failed with read timeout for "{url}"')
                    last_error = error
                else:
                    raise error

            except socket.timeout as error:
                if retry_count > 0:
                    logger_no_user_data.warning(f'HTTP {method} request failed with read timeout for "{url}"')
                last_error = error

            except http.client.IncompleteRead as error:
                logger_no_user_data.warning(
                    f'HTTP {method} request failed with incomplete read for "{url}": {repr(error)}'
                )
                last_error = error

        raise last_error or Exception(f'HTTP {method} request failed after {retries} retries for "{url}"')



import errno
import os
import stat
from datetime import datetime, timezone
from time import time

from biolib._internal.libs.fusepy import FUSE, FuseOSError, Operations
from biolib.biolib_errors import BioLibError
from biolib.jobs import Job
from biolib.typing_utils import Dict, List, Optional, Tuple, TypedDict


class _AttributeDict(TypedDict):
    st_atime: int
    st_ctime: int
    st_gid: int
    st_mode: int
    st_mtime: int
    st_nlink: int
    st_size: int
    st_uid: int


_SUCCESS_CODE = 0


class ExperimentFuseMount(Operations):
    def __init__(self, experiment):
        self._experiment = experiment
        self._job_names_map: Optional[Dict[str, Job]] = None
        self._jobs_last_fetched_at: float = 0.0
        self._mounted_at_epoch_seconds: int = int(time())

    @staticmethod
    def mount_experiment(experiment, mount_path: str) -> None:
        FUSE(
            operations=ExperimentFuseMount(experiment),
            mountpoint=mount_path,
            nothreads=True,
            foreground=True,
            allow_other=False,
        )

    def getattr(self, path: str, fh=None) -> _AttributeDict:
        if path == '/':
            return self._get_directory_attributes(timestamp_epoch_seconds=self._mounted_at_epoch_seconds)

        job, path_in_job = self._parse_path(path)
        job_finished_at_epoch_seconds: int = int(
            datetime.fromisoformat(job.to_dict()['finished_at'].rstrip('Z')).replace(tzinfo=timezone.utc).timestamp()
        )

        if path_in_job == '/':
            return self._get_directory_attributes(timestamp_epoch_seconds=job_finished_at_epoch_seconds)

        try:
            file = job.get_output_file(path_in_job)
            return self._get_file_attributes(
                timestamp_epoch_seconds=job_finished_at_epoch_seconds,
                size_in_bytes=file.length,
            )
        except BioLibError:
            # file not found
            pass

        file_paths_in_job = [file.path for file in job.list_output_files()]

        for file_path_in_job in file_paths_in_job:
            if file_path_in_job.startswith(path_in_job):
                return self._get_directory_attributes(timestamp_epoch_seconds=job_finished_at_epoch_seconds)

        raise FuseOSError(errno.ENOENT) from None  # No such file or directory

    def readdir(self, path: str, fh: int) -> List[str]:
        directory_entries = ['.', '..']

        if path == '/':
            directory_entries.extend(self._get_job_names_map(refresh_jobs=True).keys())
        else:
            job, path_in_job = self._parse_path(path)
            dir_path_in_job = '/' if path_in_job == '/' else path_in_job + '/'
            depth = dir_path_in_job.count('/')
            directory_entries.extend(
                set(
                    [
                        file.path.split('/')[depth]
                        for file in job.list_output_files()
                        if file.path.startswith(dir_path_in_job)
                    ]
                )
            )

        return directory_entries

    def open(self, path: str, flags: int) -> int:
        job, path_in_job = self._parse_path(path)
        try:
            job.get_output_file(path_in_job)
        except BioLibError:
            # file not found
            raise FuseOSError(errno.ENOENT) from None

        return 1234  # dummy file handle

    def read(self, path: str, size: int, offset: int, fh: int) -> bytes:
        job, path_in_job = self._parse_path(path)
        try:
            file = job.get_output_file(path_in_job)
        except BioLibError:
            raise FuseOSError(errno.ENOENT) from None  # No such file or directory

        return file.get_data(start=offset, length=size)

    def release(self, path: str, fh: int) -> int:
        return _SUCCESS_CODE

    def releasedir(self, path: str, fh: int) -> int:
        return _SUCCESS_CODE

    def flush(self, path: str, fh: int) -> int:
        return _SUCCESS_CODE

    @staticmethod
    def _get_directory_attributes(timestamp_epoch_seconds: int) -> _AttributeDict:
        return _AttributeDict(
            st_atime=timestamp_epoch_seconds,
            st_ctime=timestamp_epoch_seconds,
            st_gid=os.getgid(),
            st_mode=stat.S_IFDIR | 0o555,  # Directory that is readable and executable by owner, group, and others.
            st_mtime=timestamp_epoch_seconds,
            st_nlink=1,
            st_size=1,
            st_uid=os.getuid(),
        )

    @staticmethod
    def _get_file_attributes(timestamp_epoch_seconds: int, size_in_bytes: int) -> _AttributeDict:
        return _AttributeDict(
            st_atime=timestamp_epoch_seconds,
            st_ctime=timestamp_epoch_seconds,
            st_gid=os.getgid(),
            st_mode=stat.S_IFREG | 0o444,  # Regular file with read permissions for owner, group, and others.
            st_mtime=timestamp_epoch_seconds,
            st_nlink=1,
            st_size=size_in_bytes,
            st_uid=os.getuid(),
        )

    def _get_job_names_map(self, refresh_jobs=False) -> Dict[str, Job]:
        current_time = time()
        if not self._job_names_map or (current_time - self._jobs_last_fetched_at > 1 and refresh_jobs):
            self._jobs_last_fetched_at = current_time
            self._job_names_map = {job.get_name(): job for job in self._experiment.get_jobs(status='completed')}

        return self._job_names_map

    def _parse_path(self, path: str) -> Tuple[Job, str]:
        path_splitted = path.split('/')
        job_name = path_splitted[1]
        path_in_job = '/' + '/'.join(path_splitted[2:])
        job = self._get_job_names_map().get(job_name)
        if not job:
            raise FuseOSError(errno.ENOENT)  # No such file or directory

        return job, path_in_job

    # ----------------------------------- File system methods not implemented below -----------------------------------

    def chmod(self, path, mode):
        raise FuseOSError(errno.EACCES)

    def chown(self, path, uid, gid):
        raise FuseOSError(errno.EACCES)

    def mknod(self, path, mode, dev):
        raise FuseOSError(errno.EACCES)

    def rmdir(self, path):
        raise FuseOSError(errno.EACCES)

    def mkdir(self, path, mode):
        raise FuseOSError(errno.EACCES)

    def unlink(self, path):
        raise FuseOSError(errno.EACCES)

    def symlink(self, target, source):
        raise FuseOSError(errno.EACCES)

    def rename(self, old, new):
        raise FuseOSError(errno.EACCES)

    def link(self, target, source):
        raise FuseOSError(errno.EACCES)

    def utimens(self, path, times=None):
        raise FuseOSError(errno.EACCES)

    def create(self, path, mode, fi=None):
        raise FuseOSError(errno.EACCES)

    def write(self, path, data, offset, fh):
        raise FuseOSError(errno.EACCES)

    def truncate(self, path, length, fh=None):
        raise FuseOSError(errno.EACCES)

    def fsync(self, path, datasync, fh):
        raise FuseOSError(errno.EACCES)
